/**
 * 
 * Copyright 2005 LogicBlaze, Inc. http://www.logicblaze.com
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.servicemix.components.util;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.jbi.JBIException;
import javax.resource.spi.work.Work;
import javax.resource.spi.work.WorkException;
import javax.resource.spi.work.WorkManager;
import java.util.Date;
import java.util.Timer;
import java.util.TimerTask;

/**
 * An implementation inheritence class for a component which
 * polls some resource at periodic intervals to decide if there is an event to process.
 *
 * @version $Revision: 1.2 $
 */
public abstract class PollingComponentSupport extends ComponentSupport implements Work {
    private static final Log log = LogFactory.getLog(PollingComponentSupport.class);

    private WorkManager workManager;
    private Timer timer;
    private Date firstTime;
    private long period = 5000;
    private long delay;

    /**
     * Polls the underlying resource to see if some event is required
     *
     * @throws JBIException
     */
    public abstract void poll() throws Exception;

    public void release() {
    }

    public void run() {
        try {
            poll();
        }
        catch (Exception e) {
            log.error("Caught exception while polling: " + e, e);
        }
    }

    // Properties
    //-------------------------------------------------------------------------
    public WorkManager getWorkManager() {
        return workManager;
    }

    public void setWorkManager(WorkManager workManager) {
        this.workManager = workManager;
    }

    public long getDelay() {
        return delay;
    }

    public void setDelay(long delay) {
        this.delay = delay;
    }

    public Date getFirstTime() {
        return firstTime;
    }

    public void setFirstTime(Date firstTime) {
        this.firstTime = firstTime;
    }

    public long getPeriod() {
        return period;
    }

    public void setPeriod(long period) {
        this.period = period;
    }

    public Timer getTimer() {
        return timer;
    }

    public void setTimer(Timer timer) {
        this.timer = timer;
    }

    // Implementation methods
    //-------------------------------------------------------------------------
    protected void init() throws JBIException {
        if (timer == null) {
            timer = new Timer(true);
        }
        if (workManager == null) {
            throw new IllegalArgumentException("You must specify the workManager property");
        }
        super.init();

        // now lets register the work
        TimerTask timerTask = new TimerTask() {
            public void run() {
                try {
                    // lets run the work inside the JCA worker pools to ensure
                    // the threads are setup correctly when we actually do stuff
                    getWorkManager().scheduleWork(PollingComponentSupport.this);
                }
                catch (Throwable e) {
                    log.error("Failed to schedule work: " + e, e);
                }

            }
        };
        if (firstTime != null) {
            timer.scheduleAtFixedRate(timerTask, firstTime, period);
        }
        else {
            timer.scheduleAtFixedRate(timerTask, delay, period);
        }
    }
}
