/** 
 * <a href="http://servicemix.org">ServiceMix: The open source ESB</a> 
 * 
 * Copyright 2005 RAJD Consultancy Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.servicemix.jbi.messaging;
import java.util.Map;
import javax.jbi.component.Component;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import org.activemq.util.IdGenerator;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.servicemix.MessageExchangeListener;
import org.servicemix.jbi.container.ActivationSpec;
import org.servicemix.jbi.container.JBIContainer;
import org.servicemix.jbi.framework.ComponentConnector;
import org.servicemix.jbi.framework.ComponentContextImpl;
import org.servicemix.jbi.framework.LocalComponentConnector;
import EDU.oswego.cs.dl.util.concurrent.BoundedLinkedQueue;
import EDU.oswego.cs.dl.util.concurrent.ConcurrentHashMap;

/**
 * DeliveryChannel implementation
 * 
 * @version $Revision: 1.21 $
 */
public class DeliveryChannelImpl implements DeliveryChannel {
    private static final Log log = LogFactory.getLog(DeliveryChannel.class);
    private JBIContainer container;
    private ComponentContextImpl context;
    private LocalComponentConnector componentConnector;
    private BoundedLinkedQueue queue = new BoundedLinkedQueue(1000);
    private Map acks = new ConcurrentHashMap();
    private IdGenerator idGenerator = new IdGenerator();
    private MessageExchangeFactory inboundFactory;
    private MessagingStats messagingStats;
    private boolean exchangeThrottling;
    private long throttlingTimeout = 100;
    private int throttlingInterval = 1;
    private int intervalCount = 0;
    private long lastSendTime = System.currentTimeMillis();
    private long lastReceiveTime = System.currentTimeMillis();

    /**
     * Constructor
     * 
     * @param container
     * @param componentName
     */
    public DeliveryChannelImpl(JBIContainer container, String componentName) {
        this.container = container;
        this.messagingStats = new MessagingStats(componentName);
    }

    /**
     * @return size of the inbound Queue
     */
    public int getQueueSize() {
        return queue.size();
    }

    /**
     * @return the capacity of the inbound queue
     */
    public int getQueueCapacity() {
        return queue.capacity();
    }

    /**
     * Set the inbound queue capacity
     * 
     * @param value
     */
    public void setQueueCapacity(int value) {
        queue.setCapacity(value);
    }

    /**
     * close the delivery channel
     * 
     * @throws MessagingException
     */
    public void close() throws MessagingException {
    }

    /**
     * Create a message exchange factory. This factory will create exchange instances with all appropriate properties
     * set to null.
     * 
     * @return a message exchange factory
     */
    public MessageExchangeFactory createExchangeFactory() {
        MessageExchangeFactoryImpl result = createMessageExchangeFactory();
        result.setContext(context);
        ActivationSpec activationSpec = context.getActivationSpec();
        if (activationSpec != null) {
            String componentName = context.getComponentNameSpace().getName();
            // lets auto-default the container-routing information
            QName serviceName = activationSpec.getDestinationService();
            if (serviceName != null) {
                result.setServiceName(serviceName);
                log.info("default destination serviceName for " + componentName + " = " + serviceName);
            }
            QName interfaceName = activationSpec.getDestinationInterface();
            if (interfaceName != null) {
                result.setInterfaceName(interfaceName);
                log.info("default destination interfaceName for " + componentName + " = " + interfaceName);
            }
            QName operationName = activationSpec.getDestinationOperation();
            if (operationName != null) {
                result.setOperationName(operationName);
                log.info("default destination operationName for " + componentName + " = " + operationName);
            }
            String endpointName = activationSpec.getDestinationEndpoint();
            if (endpointName != null) {
                boolean endpointSet = false;
                log.info("default destination endpointName for " + componentName + " = " + endpointName);
                if (serviceName != null && endpointName != null) {
                    endpointName = endpointName.trim();
                    ServiceEndpoint[] endpoints = container.getRegistry().getEndpointsForService(serviceName);
                    if (endpoints != null) {
                        for (int i = 0;i < endpoints.length;i++) {
                            if (endpoints[i].getEndpointName().equals(endpointName)) {
                                result.setEndpoint(endpoints[i]);
                                log.info("Set default destination endpoint for " + componentName + " to "
                                        + endpoints[i]);
                                endpointSet = true;
                                break;
                            }
                        }
                    }
                }
                if (!endpointSet) {
                    log.warn("Could not find destination endpoint for " + componentName + " service(" + serviceName
                            + ") with endpointName " + endpointName);
                }
            }
        }
        return result;
    }

    /**
     * Create a message exchange factory for the given interface name.
     * 
     * @param interfaceName name of the interface for which all exchanges created by the returned factory will be set
     * @return an exchange factory that will create exchanges for the given interface; must be non-null
     */
    public MessageExchangeFactory createExchangeFactory(QName interfaceName) {
        MessageExchangeFactoryImpl result = createMessageExchangeFactory();
        result.setInterfaceName(interfaceName);
        return result;
    }

    /**
     * Create a message exchange factory for the given service name.
     * 
     * @param serviceName name of the service for which all exchanges created by the returned factory will be set
     * @return an exchange factory that will create exchanges for the given service; must be non-null
     */
    public MessageExchangeFactory createExchangeFactoryForService(QName serviceName) {
        MessageExchangeFactoryImpl result = createMessageExchangeFactory();
        result.setServiceName(serviceName);
        return result;
    }

    /**
     * Create a message exchange factory for the given endpoint.
     * 
     * @param endpoint endpoint for which all exchanges created by the returned factory will be set for
     * @return an exchange factory that will create exchanges for the given endpoint
     */
    public MessageExchangeFactory createExchangeFactory(ServiceEndpoint endpoint) {
        MessageExchangeFactoryImpl result = createMessageExchangeFactory();
        result.setEndpoint(endpoint);
        return result;
    }

    protected MessageExchangeFactoryImpl createMessageExchangeFactory() {
        MessageExchangeFactoryImpl messageExchangeFactory = new MessageExchangeFactoryImpl(idGenerator);
        messageExchangeFactory.setContext(context);
        return messageExchangeFactory;
    }

    /**
     * @return a MessageExchange - blocking call
     * @throws MessagingException
     */
    public MessageExchange accept() throws MessagingException {
        try {
            return (MessageExchange) queue.take();
        }
        catch (InterruptedException e) {
            throw new MessagingException("accept failed", e);
        }
    }

    /**
     * return a MessageExchange
     * 
     * @param timeoutMS
     * @return Message Exchange
     * @throws MessagingException
     */
    public MessageExchange accept(long timeoutMS) throws MessagingException {
        try {
            return (MessageExchange) queue.poll(timeoutMS);
        }
        catch (InterruptedException e) {
            throw new MessagingException("accept failed", e);
        }
    }

    /**
     * routes a MessageExchange
     * 
     * @param messageExchange
     * @throws MessagingException
     */
    public void send(MessageExchange messageExchange) throws MessagingException {
        MessageExchangeImpl messageExchangeImpl = (MessageExchangeImpl) messageExchange;
        if (messageExchangeImpl.isOutbound()) {
            if (exchangeThrottling) {
                if (throttlingInterval > intervalCount) {
                    intervalCount = 0;
                    try {
                        Thread.sleep(throttlingTimeout);
                    }
                    catch (InterruptedException e) {
                        log.warn("throttling failed", e);
                    }
                }
                intervalCount++;
            }
            long currentTime = System.currentTimeMillis();
            messagingStats.getOutboundExchanges().increment();
            messagingStats.getOutboundExchangeRate().addTime(currentTime - lastSendTime);
            lastSendTime = currentTime;
            messageExchangeImpl.setSourceId(componentConnector.getComponentNameSpace());
        }
        container.sendExchange(messageExchangeImpl);
    }

    /**
     * routes a MessageExchange
     * 
     * @param messageExchange
     * @return true if processed
     * @throws MessagingException
     */
    public boolean sendSync(MessageExchange messageExchange) throws MessagingException {
        MessageExchangeImpl messageExchangeImpl = (MessageExchangeImpl) messageExchange;
        AckHelper ack = new AckHelper(messageExchangeImpl);
        acks.put(messageExchange.getExchangeId(), ack);
        send(messageExchange);
        return ack.isAcked();
    }

    /**
     * routes a MessageExchange
     * 
     * @param messageExchange
     * @param timeoutMS
     * @return true if processed
     * @throws MessagingException
     */
    public boolean sendSync(MessageExchange messageExchange, long timeoutMS) throws MessagingException {
        MessageExchangeImpl messageExchangeImpl = (MessageExchangeImpl) messageExchange;
        AckHelper ack = new AckHelper(messageExchangeImpl);
        acks.put(messageExchange.getExchangeId(), ack);
        send(messageExchange);
        return ack.isAcked(timeoutMS);
    }

    /**
     * @return Returns the container.
     */
    public JBIContainer getContainer() {
        return container;
    }

    /**
     * @param container The container to set.
     */
    public void setContainer(JBIContainer container) {
        this.container = container;
    }

    /**
     * @return Returns the componentConnector.
     */
    public ComponentConnector getConnector() {
        return componentConnector;
    }

    /**
     * Set the ComponentConnector
     * 
     * @param connector context to set.
     */
    public void setConnector(LocalComponentConnector connector) {
        this.componentConnector = connector;
    }

    /**
     * Get the context
     * 
     * @return the context
     */
    public ComponentContextImpl getContext() {
        return context;
    }

    /**
     * set the context
     * 
     * @param context
     */
    public void setContext(ComponentContextImpl context) {
        this.context = context;
    }

    /**
     * Get the MessagingStats
     * 
     * @return messaging stats
     */
    public MessagingStats getMessagingStats() {
        return messagingStats;
    }

    /**
     * Is MessageExchange sender throttling enabled ?
     * 
     * @return true if throttling enabled
     */
    public boolean isExchangeThrottling() {
        return exchangeThrottling;
    }

    /**
     * Set message throttling
     * 
     * @param value
     */
    public void setExchangeThrottling(boolean value) {
        this.exchangeThrottling = value;
    }

    /**
     * Get the throttling timeout
     * 
     * @return throttling tomeout (ms)
     */
    public long getThrottlingTimeout() {
        return throttlingTimeout;
    }

    /**
     * Set the throttling timout
     * 
     * @param value (ms)
     */
    public void setThrottlingTimeout(long value) {
        throttlingTimeout = value;
    }

    /**
     * Get the interval for throttling - number of Exchanges set before the throttling timeout is applied
     * 
     * @return interval for throttling
     */
    public int getThrottlingInterval() {
        return throttlingInterval;
    }

    /**
     * Set the throttling interval number of Exchanges set before the throttling timeout is applied
     * 
     * @param value
     */
    public void setThrottlingInterval(int value) {
        throttlingInterval = value;
    }

    /**
     * Used internally for passing in a MessageExchange
     * 
     * @param packet
     * @throws MessagingException
     */
    public void processInBound(ExchangePacket packet) throws MessagingException {
        if (packet != null) {
            if (packet.isOutbound()) {
                packet.setOutbound(false);
                long currentTime = System.currentTimeMillis();
                messagingStats.getInboundExchanges().increment();
                messagingStats.getInboundExchangeRate().addTime(currentTime - lastReceiveTime);
                lastReceiveTime = currentTime;
                MessageExchangeImpl me = createInboundExchange(packet);
                me.setPacket(packet);
                Component component = ((LocalComponentConnector) componentConnector).getComponent();
                if (component != null && component instanceof MessageExchangeListener) {
                    ((MessageExchangeListener) component).onMessageExchange(me);
                }
                else {
                    try {
                        queue.put(me);
                    }
                    catch (InterruptedException e) {
                        throw new MessagingException("processInBound(" + me + ") failed", e);
                    }
                }
            }
            else {
                AckHelper ack = (AckHelper) acks.remove(packet.getExchangeId());
                if (ack != null) {
                    ack.getMessageExchange().setPacket(packet);
                    ack.done();
                }
            }
        }
    }

    /**
     * Get Inbound Factory
     * 
     * @return the inbound message factory
     */
    public MessageExchangeFactory getInboundFactory() {
        if (inboundFactory == null) {
            inboundFactory = createExchangeFactory();
        }
        return inboundFactory;
    }

    protected MessageExchangeImpl createInboundExchange(ExchangePacket packet) throws MessagingException {
        MessageExchangeImpl me = (MessageExchangeImpl) getInboundFactory().createExchange(packet.getPattern());
        me.setPacket(packet);
        return me;
    }

    /**
     * @return pretty print
     */
    public String toString() {
        return "DeliveryChannel{" + componentConnector.getComponentNameSpace() + "}";
    }
}